#!/usr/bin/env perl

use strict;
use threads;
use Getopt::Long;
use Pod::Usage;

sub install_bins;
sub code;
sub precompute;
sub call_rnafold_threads;
sub call_rnafold;
sub compute_pairwise_aligns;
sub extract_score_list;
sub compute_multiple_aligns;
sub copy;
sub SECTION;
sub system_command;
sub system_args;
sub rnasoup;

###############################################################################################
#
# Global Variables and Command Line Options
#
###############################################################################################
#TODO
my @binaries = ("rnaclustDecodeTable", "rnaclustScores2Dist.pl", "rnaclustCleanAln.pl","rnaclustAlignRange.pl", "pgma", "locarna", "mlocarna", "RNAalifold", "RNAfold","rnasoup_consMFE.pl","RNAsoup");

my $namesfile="names";      # relative to $tgtdir
my $help;
my $man;
my $multipleAligns=0;
my $tree=0;
my $start=-1;  
my $end=-1;
my $partition=0;          # Are start and end alignments given?
my $fasta_file="";
my $tgtdir="";
my $srcdir="";
my $numberThreads=1;      # by default allow just one thread for locarna and RNAfold calls
my $M=0;                  # Number of sequences
my $N=0;                  # Number of alignments
my $locarna_options="";
my $mlocarna_options="";
my $plfold_options="";
my $dponly=0;
my $dpdir="";
my $verbose=0;	 	  # Report all third party software calls?
my $printVersion=0;
my $computePartitions=0;  # Start RNAsoup immediatly afterwards
my $SoupOnly=0;           # Start RNAsoup only
my $localAlignment=0;     # Do local multiple Alignment first
my $addSequence=0;
my $bindir="";
my $localbin="";        #Set the path for a local installation of RNAclust

GetOptions(	   
	   "verbose:i"=>\$verbose,
	   "help"=> \$help,
	   "man" => \$man,
	   "tree" => \$tree,
	   "malig" => \$multipleAligns,
           "rnasoup" => \$computePartitions,
           "rnasoup-only" => \$SoupOnly,
	   "start:i" => \$start,
	   "end:i" => \$end,
	   "fasta:s" => \$fasta_file,
           "dir:s" => \$tgtdir,
           "addSequence" => \$addSequence,
           "locarna-opts=s" => \$locarna_options,
	   "mlocarna-opts=s" => \$mlocarna_options,
           "plfold-opts=s" => \$plfold_options,
           "cpu:i" => \$numberThreads,
           "dpdir:s" => \$dpdir,
           "dponly" => \$dponly,
	   "version" => \$printVersion,
           "localbin:s" => \$localbin
	   ) || pod2usage(2);

if($printVersion){ print STDOUT "$0 version 1.3\n";  exit(1); }
if($help){ pod2usage(1); }
if($man){ pod2usage(-exitstatus => 0, -verbose => 2); }
if(($fasta_file eq "" && !($tree || $multipleAligns) && (!$SoupOnly)) || $tgtdir eq ""){ pod2usage(1);  }
if(($start>-1 && $end==-1) || ($start==-1 && $end>-1)){ pod2usage(1); exit(-1); }
if($start>-1 && $end>-1){ $partition=1; }
if($multipleAligns && $partition){ pod2usage(1); exit(-1); }
if($multipleAligns && $dponly){ pod2usage(1); exit(-1); }
if($multipleAligns && $computePartitions){ $multipleAligns=0 }

if( -e $tgtdir && !$tree && !$partition && !$multipleAligns && ($dpdir eq "") && !$addSequence && !$SoupOnly) {
    print STDERR "Target directory $tgtdir already exists.\n";
    print STDERR "Please choose a different target directory (or delete it) and run again.\n";
    exit(-1);
}

###############################################################################################
#
# Main
#
###############################################################################################

if($locarna_options){
#To be extended
    if(($locarna_options =~ /output/) || ($locarna_options =~ /sequ-local/) || ($locarna_options =~ /-L/)|| ($locarna_options =~ /-P/) ){
	print "ERROR (locarna_options): Please don't change the locarna output format!\n";
	exit(-1);
    }
}
$bindir="$tgtdir/bin/";
if(!$SoupOnly){
    if(!$tree){	
	if(!$addSequence){
	    SECTION("Local copies of binaries");
	    install_bins();
	}else{
	    #Initiates the dir variables and creates an additional subfolder
	    SECTION("Add sequences to an existing tree");
	    $srcdir=$tgtdir;
	    $tgtdir="$srcdir/additionalSeqs";
	    $bindir="$srcdir/bin/";
	    mkdir $tgtdir;
	}
	
	if(!$addSequence){
	    if(!$dpdir){
		SECTION("Compute the base pair probability matrices");
		$M=precompute($fasta_file);
		if($dponly==1){ 
		    SECTION("Cheers.$M");
		    exit(1); 
		}
		$dpdir = $tgtdir."/pp";
	    }
	    else{
		SECTION("Use base pair probability matrices from directory $dpdir");
		#$N = `grep -c '>' $fasta_file`; chomp $N;  # $fasta_file only for number of sequences needed
		$M = countSeqs($fasta_file);  # $fasta_file only for number of sequences needed
	    }
	}
	else{
	    SECTION("Compute the base pair probability matrices");
	    $M=addSeq_precompute($fasta_file);
	    
	    if($dponly==1){ 
		SECTION("Cheers.");
		exit(1); 
	    }
	    
	    $dpdir = $srcdir."/pp";
	    
	    my $oldSeqs=countSeqs("$srcdir/seqs.fasta");
	    $partition=1;
	    $start=code($oldSeqs,$oldSeqs-1)+1;
	    $end=code($M,$M-1);
	}
	
	if($start!=-1 && $end!=-1){ SECTION("Compute the pairwise alignments $start - $end");  }
	else {if(!$localAlignment){SECTION("Compute the pairwise alignments");}}
	
	compute_pairwise_aligns();
	
	if($partition && !$addSequence){ 
	    SECTION("Cheers.");
	    exit(1); 
	}
    }
    
	
# Test required input files
    if(!-e "$tgtdir/names"){
	# Precomputation step without generating pairwise alignments, but generating names and seqs.fasta
	precompute($fasta_file, 1)
    }
    if(!-e "$tgtdir/paligs/" || (!-e "$tgtdir/bin/" && !$addSequence)){
	die "ERROR ($0): Target directory '$tgtdir' must contain file 'names' as well as directories 'paligs' and 'bin'.\n";
    }
    
    SECTION("Extract the score list $tgtdir");
    extract_score_list($tgtdir);
    
    if($addSequence){
	#Generates standard score and dist-list from intermediates
	copy("$srcdir/names","$tgtdir/names_all");
	system_command("cat $tgtdir/names >> $tgtdir/names_all");
	rename "$tgtdir/names_all","$tgtdir/names";
	system_command("cat $tgtdir/addSeqs.fasta >> $tgtdir/oldSeqs.fasta");
	rename "$tgtdir/oldSeqs.fasta","$tgtdir/seqs.fasta";
	
	#Move new pp-files to subfolder
	mkdir "$tgtdir/pp";
	my $oldSeqs=countSeqs("$srcdir/seqs.fasta");
	my $newSeqs=countSeqs("$tgtdir/addSeqs.fasta");

        for(my $i=$oldSeqs+1;$i<=($oldSeqs+$newSeqs); $i++){
            system_command("mv $srcdir/pp/$i $tgtdir/pp/");
        }
    }
    
    SECTION("Generate the cluster-tree");
    system_command("$bindir/pgma $tgtdir/names $tgtdir/dist-list > $tgtdir/tree");
    if($verbose){ system_command("cat $tgtdir/tree"); }
    
}
##Else-part from $SoupOnly
else{
    my $test="$tgtdir/maligs";
    $dpdir = $tgtdir."/pp";
    if(!-e $test)
    {
	$multipleAligns=1;
    }
}
if($multipleAligns || $computePartitions){
    ## Test required input files
    if(!-e "$tgtdir/seqs.fasta" || !-e "$tgtdir/tree" || !-e "$tgtdir/names" || !-e "$dpdir" || (!-e "$tgtdir/bin" && !$addSequence) || !-e "$tgtdir/paligs"){ 
	die "ERROR ($0): Target directory '$tgtdir' must contain files 'seqs.fasta', 'tree' and 'names', as well as directories 'paligs/', 'pp/' and 'bin/'.\n"; 
    }
    
    SECTION("Compute all multiple alignments");
    
    my $maligdir=compute_multiple_aligns();
    
    SECTION("Generate cleaned alignments"); 
    system_command("$bindir/rnaclustCleanAln.pl $maligdir");
    
}

if($computePartitions || $SoupOnly){
    rnasoup();
}

SECTION("Cheers.");

###############################################################################################
#
# Subs
#
###############################################################################################

sub rnasoup
{
SECTION("Start RNAsoup.");
mkdir "$tgtdir/rnasoup";

my $fasta="$tgtdir/seqs.fasta";
my $outdir="$tgtdir/rnasoup";
my $log="$tgtdir/rnasoup/LOG";

#Todo: Test for required programs!

##################### Progressive multiple alignments for all nodes within the tree #####################
SECTION("Traverse tree and copy multiple alignments" );
open(LOG,">>$log");
print LOG "============================================================\n";
print LOG "Traverse tree and copy multiple alignments\n";
print LOG "------------------------------------------------------------\n";
close(LOG);
# There is a small bug in mlocarna - it reports a strange error in case tree ends with ';'
system "perl -ne '\$_=~s/;//; print \$_;' $tgtdir/tree > $tgtdir/.tree 2>>$log";

##################### Create mfe_consensus.txt #########################################################
#SECTION("============================================================");
SECTION("Call RNAalifold --color -d2 --noLP --aln for all internal nodes with");
print "less than 500 leaves\n";
open(LOG,">>$log");
print LOG "============================================================\n";
print LOG "Call RNAalifold --color -d2 --noLP --aln for all internal nodes with\n";
print LOG "less than 500 leaves\n";					
print LOG "------------------------------------------------------------\n";
close LOG;
system_command("$bindir/rnasoup_consMFE.pl -f $fasta -d $tgtdir/maligs/ > $tgtdir/rnasoup/mfe_consensus.txt  2>>$log");

##################### Call RNAsoup #####################
#SECTION("============================================================");
SECTION("Call RNAsoup - Spot grOUPs in RNA cluster-tree");
open(LOG,">>$log");
print LOG "============================================================\n";
print LOG "Call RNAsoup - Spot grOUPs in RNA cluster-tree\n";
print LOG "------------------------------------------------------------\n";
close LOG;

system_command("perl -ne 'if(\$_!~/;/){ chomp \$_; print \$_,\";\n\";}else{ print \$_; }' $tgtdir/tree > $tgtdir/.tree 2>>$log");
system "$bindir/RNAsoup -t $tgtdir/.tree -f $fasta -m $tgtdir/rnasoup/mfe_consensus.txt -o $outdir/ 2>>$log";

##################### Convert PS pictures to PNG #####################
#print LOG "============================================================";
SECTION("Convert PS pictures to PNG (needed for SoupViewer)");
open(LOG,">>$log");
print LOG "============================================================\n";
print LOG "Convert PS pictures to PNG (needed for SoupViewer)\n";
print LOG "------------------------------------------------------------\n";
system_command("for i in $tgtdir/maligs/*.ps; do convert \${i} \${i}.png; done;");
#system_command("for i in $tgtdir/maligs/*.ps.png; do mv \${i} \${i/.ps/\"\"}; done;");
##RENAME
opendir(DIR, "$tgtdir/maligs/") or die "ERROR ($0): Could not open directory '$tgtdir/maligs/'.\n\n";
while(my $file = readdir(DIR)){
    if($file =~ /(\S*)\.ps\.png/){
	system("mv $tgtdir/maligs/$file $tgtdir/maligs/$1.png");
    }
}
close DIR;

#TODO
#my $error=system_command("grep -i error $log");
#if($error){
#    print "------------------------------------------------------------\n";
#    print "Script exit abnormally - please check $log.\n";
#    print "------------------------------------------------------------\n";
#}

#[ -n "$error" ] && { 
#echo "------------------------------------------------------------";
#echo "Script exit abnormally - please check $log.";
#echo "------------------------------------------------------------";
}

sub install_bins
{
    # If bin directory exists already --> do not copy binaries a second time!
    if(-e "$tgtdir/bin"){ return; }
	
    mkdir $tgtdir;
    mkdir "$tgtdir/bin";
    
    # Get the full path name of the target directory
    my $tgtdir_path = `dirname $tgtdir`; chomp($tgtdir_path);
    $tgtdir_path = `cd $tgtdir_path; pwd`; chomp($tgtdir_path);    
    $tgtdir = `basename $tgtdir`; chomp($tgtdir);
    $tgtdir = "$tgtdir_path/$tgtdir"; 

    for my $f (@binaries) {
	if($localbin && (($f =~ /rnaclust/) || ($f =~ /rnasoup/i) || ($f =~ /pgma/))){
	    copy("$localbin/$f", "$tgtdir/bin");
	}
	else{
	   my $t = `which $f`; chomp($t);
	   if($t=~/^\s*$/){
		die "ERROR ($0): Could not find executable of '$f'!\n";
	   }
	   copy("$t", "$tgtdir/bin");
	}
    }
    return;
}

sub code
{
    my ($n,$m)=@_;
    my $idx=0;

    for (my $i=1; $i<$n-1; $i++) {$idx+=$i;}
    $idx+=$m;

    return $idx;
}

sub addSeq_precompute
{
    my ($infile, $namesOnly)=@_;   # if $namesOnly is set to 1 no pairwise alignments are created, just the names and seqs.fasta files are created
                                   # required in case dpdir exists already and user did not generate a names file himself
    my $ppdir = "$srcdir/pp";

    local *NAMES;
    local *IN;

    # copy source-fasta file to new subfolder and append it to allSeqs.fasta
    system_command("cp $srcdir/seqs.fasta $tgtdir/oldSeqs.fasta") or die "File cannot be copied.";
    system_command("cp $tgtdir/oldSeqs.fasta $tgtdir/allSeqs.fasta") or die "File cannot be copied.";    
    system_command("cat $infile >> $tgtdir/allSeqs.fasta") or die "File cannot be copied.";
    
    my $oldSeqs=countSeqs("$srcdir/seqs.fasta");

    #If headers in infile are not unique a id is added to the name
    my $addID=checkUniqueness("$tgtdir/allSeqs.fasta",1);
    if($addID==1){
	print STDERR "WARNING ($0): Sequence names are not unique in first word (max. 42 char), but this is required to meet\n";
	print STDERR "standard of RNAclust. A unique identifier in format 'id<number>' is added as prefix to each sequence name\n";
	print STDERR "and just the first 42 characters are used.\n";
    }
     
    if(!$namesOnly){
    	mkdir $ppdir || die "ERROR ($0): Cannot make dir $ppdir\n";
    }
    
    open(IN,$infile) || die "ERROR ($0): Cannot open $infile for reading.\n";
    
    open(NAMES,">$tgtdir/$namesfile") || die "ERROR ($0): Cannot open $tgtdir/$namesfile for writing.\n";
    my $line=<IN>;
    
    ## Collect RNAfold calls and pass them to different threads 
    ## Create copy of infile: seqs.fasta - each sequence on one line
    open(OUT, ">$tgtdir/addSeqs.fasta") || die "ERROR ($0): Cannot create file '$tgtdir/addSeqs.fasta'\n";
    
    my @aSeq=();
    my @aName=();
    my @aID=();
    my $cntID=$oldSeqs+1;
    my $cntSeq=0;
    while($line) {
    	
    	## get name
    	($line =~ /^>\s*(\S+)/) || die "ERROR ($0): Expected name.\n";
    	my $name = $1;
    	# Replace non-word characters
    	$name =~ s/\W/_/g;
    	if($addID==1){
	    $name = "id$cntID"."_".$name;
	}
	
	if(length($name)>42){
	    $name=substr($name,0,42);
	}

    	print NAMES "$name\n";
    	
    	## read sequence
    	my $seq="";
    	while (defined($line=<IN>) && ($line !~ /^>/)) {
    	    if($line =~ /^\s*$/){ next; } # ignore empty lines	
    	    $line =~ s/\s//g;
    	    $seq.=$line;
    	};
    
    	$cntSeq++;
    	print OUT ">$name\n$seq\n";
    	push(@aSeq, $seq);
    	push(@aName, $name);
    	push(@aID, $cntID++);
    
    	## Call numberThread threads
    	if(!$namesOnly){
    	        if($cntSeq==$numberThreads){	
    		    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $numberThreads);
    		    @aSeq=();   # free 
    		    @aName=();
    		    @aID=();
    		    $cntSeq=0;  # start collecting again
        	}
    	}
    }
    close(OUT);
    
    ## Call the remaining threads
    if(!$namesOnly){
    	    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $#aName+1);
    }
        
    close NAMES;
    close IN;
    return $cntID-1;  # subtract 1, otherwise code $M $M-1 calculates wrong number of pairwise alignments
}

sub checkUniqueness
{
    # check if names are unique
    my ($infile, $addSeq)=@_;
    my @inHeader; 
    my @header;
    my $addID=0;
    my %HASH;

    if($addSeq){
	#newSeqs are not formated but oldSeqs are; therefore format all
	@inHeader = readpipe("grep \">\" $infile  | perl -ne \'\$header =~s\/\\W\/\_\/g; print \"\$header\n\"\'");
    }else{
	@inHeader = readpipe("grep \">\" $infile");
    }

    foreach(@inHeader){
	my $str=$_;	
	chomp($str);
	$str =~ s/(>\w+).*/$1/; 
	if(length($str)>43){
	    $str=substr($str,0,43);
	}
	push(@header,$str);
    }
    
    foreach my $key (@header) {
	$HASH{$key}++ ;
    }
    
    foreach my $key (keys %HASH) {
	if ($HASH{$key}>1){
	    $addID=1;
	    last;
	}
    }
    return $addID;
}

sub precompute
{
    my ($infile, $namesOnly)=@_;   # if $namesOnly is set to 1 no pairwise alignments are created, just the names and seqs.fasta files are created
                                   # required in case dpdir exists already and user did not generate a names file himself

    my $ppdir = "$tgtdir/pp";

    local *NAMES;
    local *IN;  
    
    #If headers in infile are not unique a id is added to the name
    my $addID=checkUniqueness($infile,0);
    if($addID==1){
	print STDERR "WARNING ($0): Sequence names are not unique in first word (max. 42 char), but this is required to meet\n";
	print STDERR "standard of RNAclust. A unique identifier in format 'id<number>' is added as prefix to each sequence name\n";
	print STDERR "and just the first 42 characters are used.\n";
    }
    
    if(!$namesOnly){
	mkdir $ppdir || die "ERROR ($0): Cannot make dir $ppdir\n";
    }
    
    open(IN,$infile) || die "ERROR ($0): Cannot open $infile for reading.\n";
    
    open(NAMES,">$tgtdir/$namesfile") || die "ERROR ($0): Cannot open $tgtdir/$namesfile for writing.\n";
    my $line=<IN>;

    ## Collect RNAfold calls and pass them to different threads 
    ## Create copy of infile: seqs.fasta - each sequence on one line
    open(OUT, ">$tgtdir/seqs.fasta") || die "ERROR ($0): Cannot create file '$tgtdir/seqs.fasta'\n";

    my @aSeq=();
    my @aName=();
    my @aID=();
    my $cntID=1;
    my $cntSeq=0;
    while($line) {
	
	## get name
	($line =~ /^>\s*(\S+)/) || die "ERROR ($0): Expected name.\n";
	my $name = $1;
	# Replace non-word characters
	$name =~ s/\W/_/g;
	if($addID==1){
	    $name = "id$cntID"."_".$name; 
	}
	
	if(length($name)>42){
	    $name=substr($name,0,42);
	}
	
	print NAMES "$name\n";
	
	## read sequence
	my $seq="";
	while (defined($line=<IN>) && ($line !~ /^>/)) {
	    if($line =~ /^\s*$/){ next; } # ignore empty lines	
	    $line =~ s/\s//g;
	    $seq.=$line;
	};

	$cntSeq++;
	print OUT ">$name\n$seq\n";
	push(@aSeq, $seq);
	push(@aName, $name);
	push(@aID, $cntID++);

	## Call numberThread threads
	if(!$namesOnly){
	        if($cntSeq==$numberThreads){	
		    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $numberThreads);
		    @aSeq=();   # free 
		    @aName=();
		    @aID=();
		    $cntSeq=0;  # start collecting again
        	}
	}
    }
    close(OUT);

    ## Call the remaining threads
    if(!$namesOnly){
	    call_rnafold_threads(\@aSeq, \@aName, \@aID, $ppdir, $#aName+1);
    }
        
    close NAMES;
    close IN;
    return $cntID-1;  # subtract 1, otherwise code $M $M-1 calculates wrong number of pairwise alignments
}

sub call_rnafold_threads
{
    my ($ref_aSeq, $ref_aName, $ref_aID, $ppdir, $nthreads) = @_;

    my @threads;

    if($nthreads == 1){  # one cpu only --> discard use of Perl threads
	call_rnafold($ref_aName->[0], $ref_aSeq->[0], $ref_aID->[0], $ppdir);
    }
    else{
	## Call threads here
	for(my $i=0; $i<$nthreads; $i++){
	    $threads[$i]=threads->create(\&call_rnafold, $ref_aName->[$i], $ref_aSeq->[$i], $ref_aID->[$i], $ppdir);
	}
	## Wait until they are ready
	for(my $i=0; $i<$nthreads; $i++){
	    my $r = $threads[$i]->join();
	}
    }
    return;
}

sub call_rnafold
{
    my ($name, $seq, $id, $ppdir) = @_;

    $name =~ s/\//_/g; 
    $name =~ s/\./_/g;
    my $pp_outfile = "$ppdir/$id";
    my $dpplot = $name."_dp.ps";
    my $dpplot2 = $name."_dp2.ps";
    my $rnaplot = $name."_ss.ps";
    # On Ubuntu systems the default shell for the system command, which is sh, does not support '-e' argument for echo
    # Thus, must ensure that bash isused --> use system with a list as argument and not as a string.
    if($locarna_options=~/--stacking/){
	# If stacking is regarded locarna expects RNAfold -p2 dotplots
	my @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAfold -p2 -d2 > /dev/null");
        if($verbose){ @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAfold -p2 -d2"); }
        system_args(@args);
	copy "$dpplot2", "$pp_outfile";
	unlink "$rnaplot"; unlink "$dpplot"; unlink "$dpplot2";
    }
    else{
	my @args;
	if(!$localAlignment){ ##TODO: vernuenftig implementieren
	    @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAfold -p -d2 > /dev/null");
	    if($verbose){ @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAfold -p -d2"); }
	}else{
	    #@args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAplfold -W 100 -L 90 -d2 > /dev/null");
	    @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAplfold -W 40 -d2 > /dev/null");
	    if($verbose){ @args = ("bash", "-c", "echo -e \"\>$name\n$seq\" | RNAplfold ..."); }
	}
        system_args(@args);
	copy "$dpplot", "$pp_outfile";
	unlink "$rnaplot"; unlink "$dpplot";
    }
    return;
}

sub compute_pairwise_aligns
{
    if(!$partition){
	$N=code($M,$M-1);
	print "  be patient, we compute $N alignments ...\n";
	system_command("$bindir/rnaclustAlignRange.pl --dpdir $dpdir --range 1-$N --locarna-path $bindir/locarna --tgtdir $tgtdir --cpu $numberThreads --verbose $verbose --locarna-opts \"$locarna_options\"");
	unlink "alirna.ps";
    }
    else{
	my $maxN=code($M,$M-1);
	if($end>$maxN){ die "ERROR ($0): '--end $end' exceeds maximal number of alignments, wich is $maxN.\n"; }
	$N=$end-$start+1;
	print "  be patient, we compute $N alignments ...\n";
	system_command("$bindir/rnaclustAlignRange.pl --dpdir $dpdir --range $start-$end --locarna-path $bindir/locarna --tgtdir $tgtdir --cpu $numberThreads --verbose $verbose --locarna-opts \"$locarna_options\"");
	unlink "alirna.ps";
	if(!$addSequence){
	    exit;
	}
    }
    return;
}

#sub extract_local_alignments{
#    print "bunzip2 ...\n";
#    open(IN,"bunzip2 -c paligs/*.bz2 |") || die "Cannot read from $!";
#    open(OUT,">localAlignments") || die "Cannot write $!";
#    
#    # @ID and Score etc. are now found on same line
#    print "generate local alignment list ...\n";
#    my $foundLocalSeqs=0;
#    my $i;
#    my @seq;
#    my @head;
#    my @head_tmp;
#    my $t1="";
#    my $t2="";
#    my $norm_score;
#    my $av_length;
#    
#
#    while(my $line=<IN>) {
#	##make uconflict-input list
#	if($line =~ /HIT/){
#	    chomp($line);
#	    @head_tmp=split(/HIT/,$line);
#	    chomp($head_tmp[1]);
#	    @head=split(/\s+/,$head_tmp[1]);
#	    #$av_length=((($head[4]-$head[2])+($head[5]-$head[3]))/2);
#	}
#	else{
#	    if($line=~/^(\d+)/){
#		if(!$t1){
#		    $t1=$1-1;
#		}else{
#		    $t2=$1-1; 
#		    #if($av_length>20){
##print OUT "$t1\t$head[3]\t$head[5]\t$t2\t$head[4]\t$head[6]\t$head[2]\n"; 
#		    print OUT "$t1\t$head[2]\t$head[4]\t$t2\t$head[3]\t$head[5]\t$head[1]\n"; 
#		    $foundLocalSeqs++;
#		    $t1=""; 
#		    $t2="";
#		    #	    }
#		}
#	    }	
#	}
#    }
#    
#    close IN;
#    close OUT;
#    if($foundLocalSeqs==0){
#	die "ERROR ($0): Failed to extract sequences!\n";
#    }
#    print "$foundLocalSeqs local sequences have been extracted\n";
#}


sub extract_score_list
{
    my ($tgtdir)=@_;
    print "bunzip2 ...\n";
    open(IN,"bunzip2 -c $tgtdir/paligs/*.bz2 |") || die "Cannot read from $!";
    open(OUT,">$tgtdir/score-list") || die "Cannot write $!";
    
    # @ID and Score etc. are now found on same line
    print "generate score list ...\n";
    my $foundScores=0;
    while(my $line=<IN>) {
	if($line =~ /^\@(\d+)/){
	    print OUT "$1 ";
	}
	#if($line =~ /Score:\s+([\-\d\.]+)\s+SCI:\s+([\d\.\-]+)\s+ID:\s+([\d\.\-]+)\s+([\d\.\-]+)\s*$/) {
	if($line =~ /Score:\s+([\-\d\.]+)\s*$/) {  # sci and identity are actually not needed --> saves computation time
	    # $1: Score $2: SCI $3: sequence identity $4: sequence identity of non-structural positions
	    print OUT "$1\n";
	    $foundScores++;
	}
    }
    close IN;
    close OUT;
    if($foundScores==0){
	die "ERROR ($0): Failed to extract Score!\n";
    }
    print "$foundScores scores have been extracted\n";
   
    print "decode table ...\n";
    system_command("$bindir/rnaclustDecodeTable < $tgtdir/score-list > $tgtdir/score-list.dec");
    
    rename "$tgtdir/score-list.dec", "$tgtdir/score-list";
    
    if($addSequence){
	copy("$srcdir/score-list","$tgtdir/score-list_all");
	system_command("cat $tgtdir/score-list >> $tgtdir/score-list_all");
	rename "$tgtdir/score-list_all","$tgtdir/score-list";
    }

    print "scores2dist ...\n";
    system_command("$bindir/rnaclustScores2Dist.pl < $tgtdir/score-list > $tgtdir/dist-list");
    
    return;
}

sub compute_multiple_aligns
{
    open(NAMES,"$tgtdir/names") || die "Cannot read 'names' file!";
    my $maligdir="$tgtdir/maligs";
    mkdir "$maligdir.mloc";
    my $i=0;
    while (my $name=<NAMES>) {
    	$i++;
    	my $short_name = sprintf "%16s",$name;
    	$short_name =~ s/[^A-Za-z0-9]/_/g;
    	system("ln -s $tgtdir/pp/$i $maligdir.mloc/$short_name");
    }
    close NAMES;
    
    
    #my $incl = `pwd`; chomp($incl);
    system_command("$bindir/mlocarna --treefile $tgtdir/tree --tgtdir $maligdir.mloc --skip-pp $mlocarna_options $tgtdir/seqs.fasta > $tgtdir/mlocarna.out");
    
    opendir(DIR,"$maligdir.mloc/intermediates");
    my @clusterfiles = readdir(DIR);
    closedir DIR;
    
    mkdir "$maligdir";
    for my $f (@clusterfiles) {
	if ($f =~ /^intermediate(\d+).aln$/) {
	    my $idx=$1;
	    rename "$maligdir.mloc/intermediates/$f","$maligdir/$idx.aln";
	}
    }

    # clean up
    system "rm -rf $maligdir.mloc";

    return $maligdir;
}

###############################################################################################
#
# Helpers
#
###############################################################################################

sub copy{ system("cp @_"); } 

 sub countSeqs
{
    my ($infile)=@_;
    
    local *IN;

    my $count=0;
    open(IN,$infile) || die "ERROR ($0): Cannot open $infile for reading.\n";
    
    foreach my $line(<IN>){
	if($line=~ /^>\s*(\S+)/){
	    $count++;
	}
    }
    close IN;
    return $count;
}
 
 
sub SECTION
{
    my ($s)=@_;
    print "============================================================\n";
    print "$s\n";
    print "------------------------------------------------------------\n";
}

sub system_command
{
    my($sysCommand) = @_;

    if($verbose){ print "CALL $sysCommand\n"; }
    system("$sysCommand")==0
        or die "ERROR ($0): Could not call system command '$sysCommand'!\n\n";
    return 1;
}

sub system_args
{
    my(@sysArgs) = @_;

    if($verbose){ print "CALL $@ysArgs\n"; }
    system(@sysArgs)==0
        or die "ERROR ($0): Could not call system command '@sysArgs'!\n\n";
    return 1;
}


###############################################################################################
#
# Docu
#
###############################################################################################

__END__

=head1 NAME

RNAclust.pl

=head1 SYNOPSIS

RNAclust.pl [options]

Options:

    --fasta	<file name>			FASTA file
						(required)

    --dir	<dir name>			Target directory TGTDIR
						(required)

    --cpu       <integer>      		   	Number of CPUs available on your machine. Calls to locarna and 
                                		RNAfold will be distributed equally on those CPUs. (default: 1)
                                                (optional)

    --dponly                                    Stop after dotplots are created. This option is useful if 
                                                calculation of pairwise alignments will be distributed among 
                                                different machines. 
                                                (optional)

    --dpdir     <dir>                           Directory of dotplots. Use this option in case you distribute the 
                                                calculation of pairwise alignments among different machines using
                                                --start and --end options.
                                                (optional)

    --start     <integer>        		Start with pairwise alignment i.
						(optional)

    --end       <integer>	        	Stop with pairwise alignment j.
						(optional, except --start is specified)

    --tree					Use this option after all pairwise alignments have been computed 
						on different machines (options --start, --end) in order to generate 
						the hierarchical cluster-tree. The pairwise alignments must be moved 
						to one 'paligs' directory beforehand.
						(optional)

    --malig					Final cluster-tree is traversed and multiple alignment for each internal 
						node is computed progressively, default: no 
						(optional)
    
    --rnasoup                                   Starts RNAsoup immediatly after computing the pairwise alignments and 
                                                the hierarchical tree. Results are written to the subfold 'rnasoup'. 
                                                Activates the --malig option.
                                                (optional)

    --rnasoup-only                              Starts just RNAsoup on an existing RNAclust directory. Activates --malig 
                                                if there are no multiple alignments available. Writes RNAsoup output to a 
                                                subfolder called 'rnasoup'.
                                                (optional)

    --addSequence	                        Appends the sequences in the fasta file to an existing cluster tree. The 
                                                folder containing the existing tree is defined by --dir. A subfolder, 
                                                named 'additionalSeqs', containing the new tree and additional files will 
                                                be created there. This function is still experimental, not all combinations 
                                                of parameter will work.
                                                (optional)
 

    --locarna-opts <"locarna options">		Options passed directly to locarna (must be given as one string)
						(optional)

    --mlocarna-opts <"mlocarna options">        Options passed directly to mlocarna (must be given as one string). 
                                                mlocarna is used if multiple alignments for the internal nodes will be 
                                                calculated (see --malig).
						(optional)

    --help	 		        	brief help message
						(optional)

    --man       	    	        	full documentation
						(optional)

    --version					print version

    --localbin <path to bin dir>	        Use the rnaclust, rnasoup and pgma binaries from a local installation 
                                                instead of the global one.
	
=head1 DESCRIPTION

    RNAclust.pl is a perl script summarizing all the single steps
    required for structural clustering of RNA structure motifs. It
    requires as input a multiple FASTA file (the sequence names should
    be unique in the first word (max. 42 char) due to requirements of
    the Vienna RNA package).  In the first step it calculates for each
    input sequence the base pair probability matrix (given as dotplot)
    of its secondary structure distribution using RNAfold from the
    Vienna RNA package. Secondly, for each pair of dotplots a
    sequence-structure alignment is calculated using LocARNA. Lastly,
    a hierarchical cluster-tree (in NEWICK format) is derived by WPGMA
    clustering of the pairwise alignment distances.

    Calculating all pairwise sequence-structure alignments is the
    bottleneck of this pipeline, although comparable fast in case
    LocARNA is used.  Hence, RNAclust.pl provides the possibility to
    distribute this step between different CPUs on your computer (see
    --cpu option). Furthermore, by using --start and --end options the
    caluculation of pairwise alignments can be distributed among
    different machines.

    As the post-processing of a large tree is problematic, you may use
    RNAsoup in order to derive those subtrees which are likely to
    define a distinct structural motif. You could start it
    automatically after computing the cluster-tree by --rnasoup or
    afterwards manually using the option --rnasoup-only.

    Invocation examples
    -------------------

    (1) RNAclust.pl --fasta FASTA --dir TARGETDIR --malig

    Runs the complete clustering pipeline, i.e. calling RNAfold -p to
    create a dotplot for each input sequence. Dotplots are written to
    TARGETDIR/pp. All pairwise alignments are calculated on one
    CPU. BZIP2 archives containing all those alignments are found in
    TARGETDIR/paligs.  The cluster-tree is derived and saved in
    TARGETDIR/tree. Lastly, the cluster-tree is traversed and for each
    internal node a multiple alignment is calculated
    progressively. Multiple alignments are saved in TARGETDIR/maligs.

    (2) RNAclust.pl --fasta FASTA --dir TARGETDIR --cpu 4

    Runs the complete clustering pipeline on the same machine but
    distributing the calculation of base pair probability matrices
    (i.e. dotplots) and the pairwise alignments between 4 CPUs of your
    machine. Does not calculate the multiple alignments for each
    internal node of the final cluster-tree.

    (3) RNAclust.pl --fasta FASTA --dir TARGETDIR --rnasoup

    Automatically starts RNAsoup after computing the newick tree. If
    you use this, the calculation of the multiple alignments is
    included. The XML-tree resulting from RNAsoup is written to
    TARGETDIR/rnasoup.

    (4) RNAclust.pl --fasta FASTA --dir TARGETDIR --cpu 4 --locarna-opts "--match=40 --mismatch=10"

    As above, but two parameters are passed to LocARNA changing the
    scoring parameters.

    (5.1) RNAclust.pl --fasta FASTA --dir TARGETDIR --dponly

    Creates only the base pair probability matrices. They are saved in
    TARGETDIR/pp.

    (5.2) RNAclust.pl --fasta FASTA --dir TARGETDIR --dpdir DIR/pp --start 1000 --end 1999

    Calculates only pairwise alignments 1000 to 1999 based on the
    dotplots provided in DIR/pp. Stops after those pairwise alignments
    are calculated.

    (5.3) RNAclust.pl --dir TARGETDIR --tree

    Calculates the final cluster-tree after pairwise alignments have
    been computed on different machines. TARGETDIR/paligs must contain
    the BZIP2 files of all pairwise alignments (collect them from the
    different machines). Furthermore TARGETDIR/names must exist.

    (5.1), (5.2) and (5.3) have to be called consecutively in order to
    distribute the calculation of pairwise alignments among different
    machines.

    For a more detailed information please read the manual.

=head1 AUTHOR

    Jan Engelhardt (jane@bioinf.uni-leipzig.de), Kristin Reiche
    (kristin@bioinf.uni-leipzig.de) and Sebastian Will
    (will@informatik.uni-freiburg.de)

=cut
