#!/usr/bin/env perl

##################################################################
# PURPOSE: Go through alignments according to cluster-tree and 
#          retrieve consensus MFEs.
##################################################################

use strict;
use Getopt::Long;
use Pod::Usage;

my $aligDir='';   # Directory aligs
my $fastaFile='';  # Fasta file 
my $help=0;

if($#ARGV<0){ pod2usage(-verbose => 1); }
GetOptions('fasta:s' => \$fastaFile,
	   'f:s' => \$fastaFile,
	   'dir:s' => \$aligDir,
	   'd:s' => \$aligDir,
	   'help'=>\$help,
	   'h'=>\$help,
	   ) or pod2usage(-verbose => 1);
if($help){ pod2usage(-verbose => 2); }

opendir(DIR, $aligDir) or die "ERROR ($0): Could not open directory '$aligDir'.\n\n";
while(my $file = readdir(DIR)){
    print "$file\n";
    if($file =~ /\d+\.clean.aln\s*$/){
#    if($file =~ /intermediate\d+\.aln\s*$/){
	
	$file = $aligDir."/".$file;

	##
	# Transform Locarna Clustalw format to general Clustalw format
	# otherwise RNAalifold and coloraln.pl are not running properly.
	my $file_new = $file."_new";
	open(IN2, $file) or die "ERROR ($0): Could not open alignment file '$file'\n\n";
	open(OUTC, ">$file_new");
	while(<IN2>){
	    if($_=~/CLUSTAL/){
		print OUTC $_;
	    }
	    elsif($_=~/\~/){
		# '~' in a name is allowed
		if($_=~/^(\S+\s+)(\S+)\s*$/){
		    my $tname = $1;  
		    my $tseq = $2;
		    $tseq =~ s/\~/\-/g;
		    print OUTC "$tname$tseq\n";
		}
		else{
		    $_ =~ s/\~/\-/g;
		    print OUTC $_;
		}
	    }
	    else{
		print OUTC $_;
	    }
	}
	close OUTC;    
	close IN2;
	system("mv $file_new $file")==0 or die "ERROR ($0): Could not move new alignment file to old alignment file!\n\n";

	print STDOUT ">$file\n";

	## Get consensus MFE
	# First get sequence IDs and check number of sequences
	open(IN2, $file) or die "ERROR ($0): Could not open alignment '$file'.\n\n";
	my %currentIDs = ();
	print "INFO\t$file\n";
	while(<IN2>){
	    if($_=~/CLUSTAL/ || $_=~/\*/ ){
		next;
	    }
	    elsif($_=~/^(\S+)\s+\S+/){
		my $name = $1;
		$currentIDs{$name}=1;
	    }
	}
	close (IN2);
	my @keys = sort {$a cmp $b} keys %currentIDs;

	if($#keys+1<=500){  # For alignments with more than 500 sequences RNAalifold does not do any prediction

	    my @output = readpipe("RNAalifold --color -d2 --noLP --aln < $file"); 
	    my $noLine = $#output+1;
	    my $consMFE = '';
	    
	    # Consensus MFE
	    if($output[1] =~ /^\S+\s+\(\s*([\d\-\.]+)\s*\=/ || $noLine != 2){
		$consMFE = $1;
	    }
	    else{
		die "ERROR ($0): Failed to retrieve consensus MFE ($file).\n\n";
	    }
	    
	    print STDOUT "n: ", $#keys+1, "\n";
	    print STDOUT "@keys \n";   # white space at end is important 
	    print STDOUT "mfe: $consMFE\n";

	    # Save secondary structure plot
	    $file =~ s/\.aln$/\.ps/;
	    system("mv alirna.ps $file")==0 or print STDERR "WARNING ($0): Could not move structure plot!\n";
	     

            # Save coloured alignment
	    $file =~ s/\.ps$/\.aln.ps/;	
	    system("mv aln.ps $file")==0 or print STDERR "WARNING ($0): Could not move coloured alignment!\n";
	    
	    ##Deprecated by Vienna RNA-Package
	    # Produce coloured alignment
	    #my $aln = $file;
	    #$aln =~ s/\.ps$/\.aln/;	
	    #my $caln = $aln;
	    #$caln .= ".ps";
	    #system("coloraln.pl -s $file $aln > $caln")==0 or print STDERR "ERROR ($0): Could not create coloured alignment plot using coloraln.pl!\n"; 

	    # SplitsTree???
	    
	}
    }
    else{
	next;
    }
}


close DIR;

#################################################################
# Documentation
#################################################################

__END__

=head1 NAME

<rnasoup_consMFE.pl> - Go through directory with alignment files intermediate*.aln and call RNAalifold. Writes the number of sequences, 
                  the sequence names and the consensus MFE to STDOUT. Requires following exterin tools: RNAalifold, coloraln.pl.
                  
=head1 SYNOPSIS

 rnasoup_consMFE.pl [options]

=head1 OPTIONS

=over 3

=item B<-f, --fasta>

Original FASTA file with sequences which have been clustered.

=item B<-d, --dir>

Directory containing intermediate*.aln files.

=item B<-h, --help>

Prints manual and exits.

=back

=head1 EXAMPLES

 # rnasoup_consMFE.pl --fasta fasta.fa --dir aligs

=head1 AUTHORS

Kristin Reiche <kristin@bioinf.uni-leipzig.de>

=cut


